/**
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.deephacks.tools4j.config.internal.core.hbase;

import java.util.Arrays;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.hadoop.conf.Configuration;
import org.deephacks.tools4j.config.internal.core.hbase.HBeanRow.HBeanBytes;
import org.deephacks.tools4j.config.internal.core.hbase.HBeanRow.HBeanRowCollector;
import org.deephacks.tools4j.config.internal.core.hbase.HBeanRow.UniqueIds;
import org.deephacks.tools4j.config.model.AbortRuntimeException;
import org.deephacks.tools4j.config.model.Bean;
import org.deephacks.tools4j.config.model.Bean.BeanId;
import org.deephacks.tools4j.config.model.BeanUtils;
import org.deephacks.tools4j.config.model.Events;
import org.deephacks.tools4j.config.spi.BeanManager;

import com.google.common.collect.Sets;

/**
 * HBase implementation of Bean Manager.
 * 
 * Any mofification to storage (create, set, merge, delete) must be validated in order 
 * to ensure row uniqueness and referential integrity between rows. 
 * 
 * Since HBase does not have the foreign keys; this is a manual process that begin by 
 * fetching involved beans, enforce validation rules and then synchronize updates back 
 * to storage if no violations are found.
 * 
 * In order to do this efficiently; rows targeted for modification are fetched in one 
 * bulk and then written back to storage in one batch. 
 * 
 * @author Kristoffer Sjogren
 */
public class HBaseBeanManager extends BeanManager {
    private static final long serialVersionUID = 2158417054421142053L;
    private final HBeanTable table;
    private final UniqueIds uids;

    public HBaseBeanManager(Configuration conf) {
        uids = HBeanBytes.createUids(conf);
        table = new HBeanTable(conf, uids);
    }

    @Override
    public void create(Bean bean) throws AbortRuntimeException {
        create(Arrays.asList(bean));
    }

    /**
     * This operation must validate the following things.
     * 
     * 1) CFG303: Parents does not already exist in storage.
     * 2) CFG301: Children exist either in storage OR memory. 
     */
    @Override
    public void create(Collection<Bean> beans) throws AbortRuntimeException {
        final HBeanRowSynchronization relations = new HBeanRowSynchronization(beans, uids);
        final Set<HBeanRow> result = table.getLazy(relations.uniqueRows());
        relations.syncStorage(result);
        relations.validateNoStorageParents();

        //        Set<HBeanRow> storageRows = new HashSet<HBeanRow>();
        //        final Set<HBeanRow> memoryRows = new HashSet<HBeanRow>();
        //        final Set<HBeanRow> refRows = new HashSet<HBeanRow>();
        //        for (Bean target : beans) {
        //            final HBeanRow t = new HBeanRow(target, uids);
        //            storageRows.add(t);
        //            memoryRows.add(t);
        //            for (BeanId id : target.getReferences()) {
        //                HBeanRow r = new HBeanRow(id, uids);
        //                refRows.add(r);
        //                storageRows.add(r);
        //            }
        //        }
        //        storageRows = table.getLazy(storageRows, FetchType.KEY_ONLY);
        //        for (HBeanRow row : storageRows) {
        //            // ensure beans to be create does not exist
        //            if (memoryRows.contains(row)) {
        //                throw Events.CFG303_BEAN_ALREADY_EXIST(row.getBeanId());
        //            }
        //        }
        //        // ensure references exist either in memory or storage
        //        for (HBeanRow row : refRows) {
        //            if (!memoryRows.contains(row) && !storageRows.contains(row)) {
        //                throw Events.CFG301_MISSING_RUNTIME_REF(row.getBeanId());
        //            }
        //        }
        //        // select the rows (memory and storage) to update predecessors
        //        final Set<HBeanRow> toUpdate = new HashSet<HBeanRow>();
        //        for (HBeanRow memory : refRows) {
        //            for (HBeanRow row : memoryRows) {
        //                if (memory.equals(row)) {
        //                    toUpdate.add(row);
        //                }
        //            }
        //            for (HBeanRow row : storageRows) {
        //                if (memory.equals(row)) {
        //                    toUpdate.add(row);
        //                }
        //            }
        //
        //        }
        //
        //        table.put(memoryRows);
        //        table.put(toUpdate);

    }

    @Override
    public void createSingleton(BeanId singleton) {
        return;
    }

    @Override
    public void set(Bean bean) {
        set(Arrays.asList(bean));
    }

    /**
     * This operation must validate the following things.
     * 
     * 1) CFG304: Parents exist in storage.
     * 2) CFG301: Children exist in storage. 
     * 
     */
    @Override
    public void set(Collection<Bean> beans) throws AbortRuntimeException {
        Set<HBeanRow> rows = new HashSet<HBeanRow>();
        for (Bean bean : beans) {
            rows.add(new HBeanRow(bean, uids));
        }
        Set<HBeanRow> result = table.getLazy(rows);
        for (Bean bean : beans) {
            HBeanRow id = new HBeanRow(bean.getId(), uids);
            if (!result.contains(id)) {
                throw Events.CFG304_BEAN_DOESNT_EXIST(bean.getId());
            }
        }
        // it seems hbase 'put' operations ignore nullifying qualifiers but
        // surely there must be a better way than to delete then create
        table.delete(rows);
        table.put(rows);
    }

    @Override
    public void merge(Bean bean) throws AbortRuntimeException {
        merge(Arrays.asList(bean));
    }

    /**
     * This operation must validate the following things.
     * 
     * 1) CFG304: Parents exist in storage.
     * 2) CFG301: Children exist in storage. 
     * 
     */
    @Override
    public void merge(Collection<Bean> beans) throws AbortRuntimeException {
        Set<HBeanRow> rows = new HashSet<HBeanRow>();
        for (Bean bean : beans) {
            rows.add(new HBeanRow(bean.getId(), uids));
        }
        Set<HBeanRow> result = table.getLazy(rows);
        for (Bean bean : beans) {
            HBeanRow id = new HBeanRow(bean.getId(), uids);
            if (!result.contains(id)) {
                throw Events.CFG304_BEAN_DOESNT_EXIST(bean.getId());
            }
        }
        for (HBeanRow stored : result) {
            for (Bean merge : beans) {
                if (stored.getBeanId().equals(merge.getId())) {
                    stored.merge(merge);
                }
            }
        }
        table.put(result);
    }

    @Override
    public Bean getEager(BeanId id) throws AbortRuntimeException {
        final HashSet<HBeanRow> row = Sets.newHashSet(new HBeanRow(id, uids));
        HBeanRowCollector result = table.getEager(row);
        List<Bean> beans = result.getBeans();
        if (beans == null || beans.size() == 0) {
            throw Events.CFG304_BEAN_DOESNT_EXIST(id);
        }
        return beans.get(0);
    }

    @Override
    public Bean getLazy(BeanId id) throws AbortRuntimeException {
        final HashSet<HBeanRow> row = Sets.newHashSet(new HBeanRow(id, uids));
        Set<HBeanRow> result = table.getLazy(row);
        if (result == null || result.size() == 0) {
            throw Events.CFG304_BEAN_DOESNT_EXIST(id);
        }
        return result.iterator().next().getBean();
    }

    @Override
    public Bean getSingleton(String schemaName) throws IllegalArgumentException {
        return list(schemaName).get(0);
    }

    @Override
    public Map<BeanId, Bean> list(String schemaName) throws AbortRuntimeException {
        HBeanRowCollector result = table.listEager(schemaName);
        List<Bean> beans = result.getBeans();
        return BeanUtils.uniqueIndex(beans);
    }

    @Override
    public void delete(BeanId id) throws AbortRuntimeException {
        delete(id.getSchemaName(), Arrays.asList(id.getInstanceId()));
    }

    /**
     * This operation must validate the following things.
     * 
     * 1) CFG302: No parents reference these beans. 
     * 
     */
    @Override
    public void delete(String schemaName, Collection<String> instanceIds)
            throws AbortRuntimeException {
        Set<HBeanRow> rows = new HashSet<HBeanRow>();
        for (String id : instanceIds) {
            rows.add(new HBeanRow(BeanId.create(id, schemaName), uids));
        }
        table.delete(rows);
    }

}
