/**
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.deephacks.tools4j.config.internal.core.hbase;

import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Set;

import org.deephacks.tools4j.config.internal.core.hbase.HBeanRow.HBeanBytes;
import org.deephacks.tools4j.config.internal.core.hbase.HBeanRow.UniqueIds;
import org.deephacks.tools4j.config.model.Bean;
import org.deephacks.tools4j.config.model.Bean.BeanId;
import org.deephacks.tools4j.config.model.Events;

import com.google.common.collect.ArrayListMultimap;
import com.google.common.collect.Multimap;

/**
 * HBase does not have the foreign keys so any mofification to storage (create, 
 * set, merge, delete) must be validated in order to ensure uniqueness and referential 
 * integrity between rows. 
 * 
 * This is a manual process that begin by fetching involved beans, enforce 
 * validation rules and then synchronize updates back to storage if no violations 
 * are found.
 * 
 * In order to do this efficiently; rows targeted for modification are fetched in one 
 * bulk and then written back to storage in one batch. 
 * 
 */
public class HBeanRowSynchronization {
    private final Multimap<Long, HBeanRow> memoryPtoC = ArrayListMultimap.create();
    private final Multimap<Long, HBeanRow> memoryCtoP = ArrayListMultimap.create();

    private final Multimap<Long, HBeanRow> storagePtoC = ArrayListMultimap.create();
    private final Multimap<Long, HBeanRow> storageCtoP = ArrayListMultimap.create();
    private final UniqueIds uids;

    public HBeanRowSynchronization(Collection<Bean> beans, UniqueIds uids) {
        this.uids = uids;
        for (Bean bean : beans) {
            HBeanRow parent = new HBeanRow(bean, uids);
            for (BeanId id : bean.getReferences()) {
                addMemory(parent, new HBeanRow(id, uids));
            }
        }
    }

    /**
     * Add parent and child to this synchronization object. 
     */
    private void addMemory(HBeanRow parent, HBeanRow child) {
        memoryPtoC.put(parent.getId(), child);
        memoryCtoP.put(child.getId(), parent);
    }

    /**
     * @return A unique set of in-memory rows of parents and children.
     */
    public Set<HBeanRow> uniqueRows() {
        HashSet<HBeanRow> rows = new HashSet<HBeanRow>();
        rows.addAll(memoryCtoP.values());
        rows.addAll(memoryPtoC.values());
        return rows;
    }

    /**
     * Sync a set of storage rows towards memory representation.      
     * 
     * @param storageRows
     */
    @SuppressWarnings("unchecked")
    public void syncStorage(Set<HBeanRow> storageRows) {
        for (HBeanRow row : storageRows) {
            if (isParent(row)) {
                storagePtoC.putAll(row.getId(), Collections.EMPTY_LIST);
            } else if (isChild(row)) {
                for (HBeanRow parent : memoryCtoP.get(row.getId())) {
                    storagePtoC.put(parent.getId(), row);
                }
            } else {
                throw new IllegalArgumentException("Row [" + row + "] is not parent or child.");
            }
        }
    }

    public Collection<HBeanRow> getChildren(HBeanRow parent) {
        if (!isParent(parent)) {
            throw new IllegalArgumentException("Row " + parent + " is not a parent.");
        }
        return memoryPtoC.get(parent.getId());
    }

    public Collection<HBeanRow> getParents(HBeanRow child) {
        if (!isChild(child)) {
            throw new IllegalArgumentException("Row " + child + " is not a child.");
        }
        return memoryPtoC.get(child.getId());
    }

    /**
     * Check if provided (storage) row is a child.  
     */
    public boolean isChild(HBeanRow storageRow) {
        return memoryPtoC.get(storageRow.getId()) != null ? true : false;
    }

    /**
     * Check if provided (storage) row is a parent.  
     */
    public boolean isParent(HBeanRow storageRow) {
        return memoryCtoP.get(storageRow.getId()) != null ? true : false;
    }

    /**
     * Validate that no parent beans exist as rows in storage (has same rowkey).
     * 
     * @param storageRows rows currently in storage. 
     */
    public void validateNoStorageParents() {
        for (Long parent : storagePtoC.keys()) {
            throw Events.CFG303_BEAN_ALREADY_EXIST(HBeanBytes.getBeanId(parent, uids));
        }
    }

}
