/**
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.deephacks.tools4j.config.internal.core.hbase;

import java.util.Collection;
import java.util.HashSet;
import java.util.Set;

import org.deephacks.tools4j.config.internal.core.hbase.HBeanRow.UniqueIds;
import org.deephacks.tools4j.config.internal.core.hbase.HBeanTable.FetchType;
import org.deephacks.tools4j.config.model.Bean;
import org.deephacks.tools4j.config.model.Events;

import com.google.common.collect.ArrayListMultimap;
import com.google.common.collect.Multimap;

/**
 * Class for managing HBeanRows when doing create, set and merge operations.
 */
public class HBeanRows {
    /** User provided beans located in-memory; targeted for create, set, merge */
    private final Multimap<HBeanRow, HBeanRow> memory;
    /** HBeanRow (key) and references (value) storage representation of user provided beans */
    private final Multimap<HBeanRow, HBeanRow> storage;
    /** table that store beans */
    private final HBeanTable table;
    /** id lookup facility */
    private final UniqueIds uids;

    /**
     * Fetch beans + their bean references and initalize them according 
     * to storage. If any bean fails to be found an exceptions will be thrown.
     */
    private HBeanRows(final Multimap<HBeanRow, HBeanRow> memory,
            final Multimap<HBeanRow, HBeanRow> storage, final HBeanTable table, final UniqueIds uids) {
        this.table = table;
        this.uids = uids;
        this.storage = storage;
        this.memory = memory;
    }

    public static HBeanRows prepareCreate(Collection<Bean> beans, HBeanTable table, UniqueIds uids) {
        // beans and references are fetched in one single query
        // we must rearrange the result so that relationship between
        // beans and references are restored
        final HBeanRowSynchronization relations = new HBeanRowSynchronization(beans, uids);
        final Multimap<HBeanRow, HBeanRow> storage = ArrayListMultimap.create();
        final Set<HBeanRow> result = table.getLazy(relations.uniqueRows());

        for (HBeanRow row : result) {
            if (relations.isParent(row)) {
                throw Events.CFG303_BEAN_ALREADY_EXIST(row.getBeanId());
            } else if (relations.isChild(row)) {
                for (HBeanRow parent : relations.getParents(row)) {
                    storage.put(parent, row);
                }
            } else {
                throw new IllegalStateException("Incorrectly created relations or corrupt storage.");
            }
        }
        return null;
//        return new HBeanRows(memory, null, table, uids);
    }

    public static HBeanRows prepareSet(Collection<Bean> beans, HBeanTable table, UniqueIds uids) {
    	return null;

    }

    public static HBeanRows prepareMerge(Collection<Bean> beans, HBeanTable table, UniqueIds uids) {
    	return null;
    }

    /**
     * Return the HBeanRow references of a particular row. May be null. 
     * 
     * @param row the parent HBeanRow
     * @return A list of HBeanRow references.
     */
    public Set<HBeanRow> getReferences(HBeanRow row) {
        return new HashSet<HBeanRow>(memory.get(row));
    }

    public void ensureTargetNotExist() {
        final Set<HBeanRow> result = table.getLazy(memory.keySet(), FetchType.KEY_ONLY);
        if (result.size() > 0) {
            HBeanRow row = result.iterator().next();
            throw Events.CFG303_BEAN_ALREADY_EXIST(row.getBeanId());
        }
    }

    /**
     * Put the references.
     */
    public void putReferences() {

    }

    public void updateReferences() {

    }

    public void putTargets() {
//        table.put(targets.keySet());
    }

}
