/**
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.deephacks.tools4j.config.internal.core.hbase;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.hbase.KeyValue;
import org.apache.hadoop.hbase.client.Delete;
import org.apache.hadoop.hbase.client.Get;
import org.apache.hadoop.hbase.client.HTable;
import org.apache.hadoop.hbase.client.Put;
import org.apache.hadoop.hbase.client.Result;
import org.apache.hadoop.hbase.client.ResultScanner;
import org.apache.hadoop.hbase.client.Row;
import org.apache.hadoop.hbase.client.Scan;
import org.apache.hadoop.hbase.filter.FilterList;
import org.apache.hadoop.hbase.filter.PrefixFilter;
import org.deephacks.tools4j.config.internal.core.hbase.HBeanRow.HBeanBytes;
import org.deephacks.tools4j.config.internal.core.hbase.HBeanRow.HBeanRowCollector;
import org.deephacks.tools4j.config.internal.core.hbase.HBeanRow.UniqueIds;

/**
 * The table where beans are stored. 
 * 
 * Each bean is kept in a separate row. Each operation on a bean will be atomic on 
 * individual beans, but not accross beans. Each bean will also be versioned 
 * as a whole.   
 * 
 */
public class HBeanTable {
    /** Lookup table for short ids: sid, iid, pid. */
    private final UniqueIds uids;
    /** table used for storing key values */
    private final HTable table;
    /** the max depth allowed for traversing and fetching references */
    private final static int FETCH_DEPTH_MAX = 10;

    public HBeanTable(Configuration conf, UniqueIds uids) {
        table = HBeanBytes.getBeanTable(conf);
        this.uids = uids;
    }

    /**
     * Fetch all rows for a particular schema and traverse and fetch references eagerly.
     * 
     * @param schemaName schema to fetch
     * @param fetchType data to fetch
     * @return collector carring the result from the query
     */
    public HBeanRowCollector listEager(String schemaName, FetchType... fetchType) {
        Set<HBeanRow> rows = listLazy(schemaName, fetchType);
        HBeanRowCollector collector = new HBeanRowCollector(rows);
        getEager(rows, collector, FETCH_DEPTH_MAX, fetchType);
        return collector;
    }

    /**
     * Fetch a set of rows and traverse and fetch references eagerly.
     * 
     * @param rows to fetch
     * @param fetchType data to fetch
     * @return collector carring the result from the query
     */
    public HBeanRowCollector getEager(Set<HBeanRow> rows, FetchType... fetchType) {
        Set<HBeanRow> result = getLazy(rows, fetchType);
        HBeanRowCollector collector = new HBeanRowCollector(result);
        getEager(result, collector, FETCH_DEPTH_MAX, fetchType);
        return collector;
    }

    /**
     * Fetch a set of rows without fetching their references.
     * 
     * @param schemaName schema to fetch
     * @param fetchType data to fetch
     * @return rows found
     */
    public Set<HBeanRow> listLazy(String schemaName, FetchType... fetchType) {
        byte[] sid = HBeanBytes.extractSidPrefix(schemaName, uids);
        Scan scan = new Scan();
        HBeanBytes.setColumns(scan, fetchType);
        FilterList list = new FilterList();
        if (scan.getFilter() != null) {
            list.addFilter(scan.getFilter());
        }
        list.addFilter(new PrefixFilter(sid));
        scan.setFilter(list);
        Set<HBeanRow> rows = new HashSet<HBeanRow>();
        try {
            ResultScanner scanner = table.getScanner(scan);
            for (Result r : scanner) {
                HBeanRow row = new HBeanRow(r.raw(), uids);
                rows.add(row);
            }
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
        return rows;
    }

    /**
     * Fetch a set of rows without fetching their references.
     * 
     * @param rows  rows to fetch
     * @param fetchType data to fetch
     * @return rows found
     */
    public Set<HBeanRow> getLazy(Set<HBeanRow> rows, FetchType... fetchType) {
        final List<Row> gets = new ArrayList<Row>(rows.size());
        for (HBeanRow row : rows) {
            final Get get = new Get(row.getRowKey());
            HBeanBytes.setColumns(get, fetchType);
            gets.add(get);
        }
        Set<HBeanRow> result = new HashSet<HBeanRow>();
        try {
            final Result[] r = new Result[gets.size()];
            table.batch(gets, r);
            for (int i = 0; i < r.length; i++) {
                KeyValue[] kvs = r[i].raw();
                if (kvs.length != 0) {
                    HBeanRow row = new HBeanRow(kvs, uids);
                    result.add(row);
                }
            }
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
        return result;
    }

    /**
     * Put a set of rows.
     * 
     * @param rows to put.
     */
    public void put(Set<HBeanRow> rows) {
        final List<Row> create = new ArrayList<Row>();
        try {
            for (HBeanRow row : rows) {
                final Put write = new Put(row.getRowKey());
                write.add(row.getProperties());
                for (KeyValue kv : row.getPredecessors()) {
                    write.add(kv);
                }
                for (KeyValue kv : row.getReferences()) {
                    write.add(kv);
                }
                // hbase cannot have rowkeys without columns so we need 
                // a dummy value to represent beans without any values
                write.add(new KeyValue(row.getRowKey(), HBeanRow.DUMMY_COLUMN_FAMILY,
                        HBeanRow.DUMMY_COLUMN_FAMILY, new byte[] { 1 }));
                create.add(write);
            }
            table.batch(create);
            table.flushCommits();
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    /**
     * Recursive method for traversing and collecting row references.  
     */
    private void getEager(Set<HBeanRow> rows, HBeanRowCollector collector, int level,
            FetchType... fetchType) {
        int size = rows.size();
        if (size == 0) {
            return;
        }
        if (--level < 0) {
            return;
        }
        Set<HBeanRow> refs = new HashSet<HBeanRow>();
        for (HBeanRow row : rows) {
            refs.addAll(row.getReferenceRows());
        }
        refs = getLazy(refs, fetchType);
        // only recurse rowkeys we havent already 
        // visited to break circular references
        Set<HBeanRow> unvisited = collector.filterUnvisted(refs);
        collector.addReferences(refs);
        getEager(unvisited, collector, level, fetchType);
    }

    public static enum FetchType {
        KEY_ONLY;
    }

    /**
     * Delete a set of rows.
     * 
     * @param rows to be deleted.
     */
    public void delete(Set<HBeanRow> rows) {
        final List<Row> delete = new ArrayList<Row>();
        try {
            for (HBeanRow row : rows) {
                delete.add(new Delete(row.getRowKey()));
            }
            table.batch(delete);
            table.flushCommits();
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }
}